//////////////////////////////////////////////////////////////////////
//
//  ERRORSYS.PRG
//
//  Copyright:
//      Alaska Software, (c) 1997-2003. All rights reserved.
//
//  Contents:
//      Install default error code block
//
//  Remarks:
//      Function ErrorSys() is always called at program startup
//
//////////////////////////////////////////////////////////////////////

#Include "Xbp.ch"
#Include "Error.ch"

//#Define  DEBUG
***********************************
* Name of error log without extension
***********************************
#Define EHS_ERRORLOG "XPPERROR"

/*
 *  language specific string constant which are used in the error handler
 */
#Define EHS_CANCEL            "Cancel"
#Define EHS_EXIT_WITH_LOG     "Exit with LOG file"
#Define EHS_RETRY             "Retry"
#Define EHS_IGNORE            "Ignore"
#Define EHS_OS_ERROR          ";Operating system error : "
#Define EHS_CALLED_FROM       "Called from"
#Define EHS_XPP_ERROR_MESSAGE "Xbase++ Error Message"
#Define EHS_ERROR             "Error "
#Define EHS_WARNING           "Warning "
#Define EHS_DESCRIPTION       ";Description : "
#Define EHS_FILE              ";File : "
#Define EHS_OPERATION         ";Operation : "
#Define EHS_LOG_OPEN_FAILED   "Unable to open error log file"
#Define EHS_ERROR_LOG_OF      "ERROR LOG of "
#Define EHS_DATE              " Date:"
#Define EHS_XPP_VERSION       "Xbase++ version     :"
#Define EHS_OS_VERSION        "Operating system    :"
#Define EHS_LOG_WRITTEN_TO(cFile) "Error log was written to the file "+ cFile

*************
Proc ErrorSys()                             // Install default
*************

Begin Sequence
   ErrorBlock({|o| StandardEH(o)})             // error code block
End

Return

*************************************
* Default error handler function
*************************************
STATIC FUNCTION StandardEH(oError)
   LOCAL i, cMessage, aOptions, nOption, nSeverity
   LOCAL row, col
   LOCAL oDacSession, oSession

   /* Check If error is handled automatically */
   DO CASE

   /* Division by zero results in 0 */
   CASE oError:genCode == XPP_ERR_ZERODIV
      Return 0

   /* Zero divide by zero is also 0 and may occur
    * in operations like /, /=, % and %=
    */
   CASE oError:genCode == XPP_ERR_NUMERR
      If "/" $ oError:operation .Or. "%" $ oError:operation
          If oError:args[-1] == 0
             Return 0
          EndIf
      EndIf

   /* Error opening a file on a network */
   CASE oError:genCode == XPP_ERR_OPEN  .And. ;
        oError:osCode  == 32            .And. ;
        oError:canDefault
      Return(.F.)

   /* No lock is set */
   CASE oError:genCode == XPP_ERR_APPENDLOCK .And. ;
        oError:canDefault
      Return(.F.)

   ENDCASE

   oSession:=DbSession()
   If oSession = NIL .And. IsFunction("DacSession")
      oDacSession:=&("DacSession()")
      oSession:=oDacSession:getDefault()
   EndIf
   If oSession != NIL
       If oSession:getLastError() != 0
           oError:cargo:={oError:cargo, ;
                            oSession:getLastError(),;
                            oSession:getLastMessage() }
       EndIf
   EndIf

   /* No default handling defined: create error message */
   cMessage:=ErrorMessage(oError)

   /* Array for selection */
#ifdef DEBUG
   aOptions:={ EHS_CANCEL, EHS_EXIT_WITH_LOG }
#Else
   aOptions:={ EHS_CANCEL }
#EndIf

   If oError:canRetry
      AAdd(aOptions, EHS_RETRY)
   EndIf

   If oError:canDefault
      AAdd(aOptions, EHS_IGNORE)
   EndIf

   If ! Empty(oError:osCode)

      #IfDef XPPVER

         #If XPPVER>01810000
            cMessage+=EHS_OS_ERROR+LTrim(Str(oError:osCode))+";"+DosErrorMessage(oError:osCode)
         #Else
            cMessage+=EHS_OS_ERROR+LTrim(Str(oError:osCode))
         #EndIf

      #EndIf


   EndIf

   If AppType()<>APPTYPE_PM

      /* Display Alert() Box possible ? */
      If SetAppWindow() != NIL
         i  :=0
         row:=Row()
         col:=Col()
         DO WHILE i == 0
            i:=Alert(cMessage, aOptions)
         ENDDO
         SetPos(row, col)

         /* Perform selected option */
         If ! Empty(i)
       DO CASE
       CASE aOptions[i] == EHS_IGNORE
          Return .F.
       CASE aOptions[i] == EHS_RETRY
          Return .T.
       CASE aOptions[i] == EHS_CANCEL
          Break(oError)
       CASE aOptions[i] == EHS_EXIT_WITH_LOG
          ErrorLog(oError, 2)
       ENDCASE
    EndIf
      Else
         /* There is no console window */
#ifdef DEBUG
    ErrorLog(oError, 2)
#EndIf
      EndIf

      /*
       * Program can not or should not be continued
       * Set error level and terminate program !
       */

      ErrorLevel(1)
      QUIT
   EndIf

   If oError:canDefault .And. oError:canRetry
      nOption:=XBPMB_ABORTRETRYIGNORE
   ElseIf oError:canRetry
      nOption:=XBPMB_RETRYCANCEL
   ElseIf oError:canDefault
      nOption:=XBPMB_OKCANCEL
   Else
      nOption:=XBPMB_CANCEL
   EndIf

   i:=1
   DO WHILE ! Empty(ProcName(++i))
      cMessage+= ";" + EHS_CALLED_FROM +" " + Trim(ProcName(i))   + "(" + ;
                             LTrim(Str(ProcLine(i))) + ")"
   ENDDO

   i:=0
   /* select icon for ConfirmBox() */
   DO CASE
      CASE oError:severity == XPP_ES_FATAL
           nSeverity:=XBPMB_CRITICAL
      CASE oError:severity == XPP_ES_ERROR
           nSeverity:=XBPMB_CRITICAL
      CASE oError:severity == XPP_ES_WARNING
           nSeverity:=XBPMB_WARNING
      OTHERWISE
           nSeverity:=XBPMB_INFORMATION
   ENDCASE
   /* Display ConfirmBox() */
   i:=ConfirmBox(, StrTran(cMessage, ";", Chr(13)), ;
                    EHS_XPP_ERROR_MESSAGE , ;
                    nOption , ;
                    nSeverity + XBPMB_APPMODAL+XBPMB_MOVEABLE)

   DO CASE
   CASE i == XBPMB_RET_RETRY
      Return (.T.)
   CASE i == XBPMB_RET_IGNORE
      Return (.F.)
   CASE i == XBPMB_RET_CANCEL
#ifdef DEBUG
         If ConfirmBox(, EHS_EXIT_WITH_LOG, EHS_XPP_ERROR_MESSAGE, XBPMB_YESNO,;
                  XBPMB_WARNING+XBPMB_APPMODAL+XBPMB_MOVEABLE) != XBPMB_RET_YES
              Break(oError)
         EndIf
#Else
         Break(oError)
#EndIf
   ENDCASE

#ifdef DEBUG
   ErrorLog(oError, 2)
#EndIf

   /* Set error level and terminate program ! */
   ErrorLevel(1)
   QUIT

Return .F. /* The compiler expects a Return value */



***************************************
STATIC FUNCTION ErrorMessage(oError)
*
*  Creates a string with the important Informations
*  from the error object
***************************************

   /* Check If this is an error or warning message */
   LOCAL cMessage:=;
         IIf(oError:severity > XPP_ES_WARNING, ;
                          EHS_ERROR, EHS_WARNING)

   /* Add name of subsystem or 'unkown subsytem' */
   If Valtype(oError:subSystem) == "C"
      cMessage += oError:subSystem
   Else
      cMessage += "????"
   EndIf

   /* Add error code of subsystem */
   If Valtype(oError:subCode) == "N"
      cMessage += "/"+ LTrim(Str(oError:subCode))
   Else
      cMessage += "/????"
   EndIf

   /* Optional: Add error description */
   If Valtype(oError:description) == "C"
      cMessage += EHS_DESCRIPTION + ;
                   oError:description
   EndIf

   /* Optional: Add name of the file which were the error occured */
   If ! Empty(oError:fileName)
      cMessage += EHS_FILE + oError:fileName
   EndIf

   /* Optional: Add name of the operation which caused the error */
   If ! Empty(oError:operation)
      cMessage += EHS_OPERATION + oError:operation
   EndIf

   /* Add Thread ID of the thread on which the error occured */
   cMessage += ";Thread ID : " + ;
                LTrim(Str(oError:thread))

   If Valtype(oError:cargo)="A" .And. len(oError:cargo) == 3
      If ValType(oError:cargo[1])=="C"
         cMessage += ";" +  LineSplit(oError:cargo[1], 50)
      EndIf
      cMessage += ";" +  LineSplit(oError:cargo[3], 50)
   EndIf
Return cMessage



************************************************
STATIC PROCEDURE ErrorLog(oError, nStackStart)
*
*  Creates a string with the important Informations
*  from the error object
************************************************
   LOCAL i:=0, bError:=ErrorBlock({|e| Break(e)})
   LOCAL cErrorLog
   LOCAL cExtension:="LOG"
   LOCAL lPrint, lConsole, cAlternate, lAlternate, lExtra
   /* Save current printer related settings, turn printer off */
   lPrint    :=Set(_SET_PRINTER)
   lConsole  :=Set(_SET_CONSOLE)
   cAlternate:=Set(_SET_ALTFILE)
   lAlternate:=Set(_SET_ALTERNATE)
   lExtra    :=Set(_SET_EXTRA, .F.)

   SET PRINTER OFF
   SET CONSOLE ON

   If SetAppWindow() == NIL ;
     .Or. .NOT. SetAppWindow():isDerivedFrom(RootCrt())
      SET CONSOLE OFF
   EndIf

   /* Trap errors that might occur while opening the ALTERNATE file */
   DO WHILE .T.
      cErrorLog:=EHS_ERRORLOG + "." + cExtension
      BEGIN SEQUENCE
        SET ALTERNATE TO (cErrorLog)
        SET ALTERNATE ON
      RECOVER
        /*
         * ALTERNATE file could not be opened:
         * try other filename
         */
        cExtension:=PadL(++i,3,"0")
        If i > 999
           If AppType()<>APPTYPE_PM
              TONE(660,5)
              ? EHS_LOG_OPEN_FAILED
              If !Set(_SET_CONSOLE)
                  OutErr(CHR(10)+CHR(13) + EHS_LOG_OPEN_FAILED)
              EndIf
           Else
              MsgBox(EHS_LOG_OPEN_FAILED)
           EndIf
           ErrorLevel(1)
           QUIT
        EndIf
        Loop
      END SEQUENCE
      EXIT
   ENDDO

   ErrorBlock(bError)

   ? Replicate("-", 78)
   ? EHS_ERROR_LOG_OF +'"'+ appName(.T.) +'"'+ EHS_DATE, Date(), Time()
   ?
   ? EHS_XPP_VERSION , Version()+"."+Version(3)
   ? EHS_OS_VERSION  , Os()
   ? Replicate("-", 78)
   ? "oError:args         :"
   If Valtype(oError:Args)=="A"
      AEval(oError:Args, ;
             {|x,y| Qout(Space(9),"-> VALTYPE:", y:=Valtype(x))  , ;
                     IIf(y=="O", QQout(" CLASS:", x:className()), ;
                                  QQout(" VALUE:", Var2Char(x))) })
   Else
      Qout(Space(10),"-> NIL")
   EndIf

   ? "oError:canDefault   :" , oError:canDefault
   ? "oError:canRetry     :" , oError:canRetry
   ? "oError:canSubstitute:" , oError:canSubstitute
   ? "oError:cargo        :" , oError:cargo
   ? "oError:description  :" , oError:description
   ? "oError:filename     :" , oError:filename
   ? "oError:genCode      :" , oError:genCode
   ? "oError:operation    :" , oError:operation
   ? "oError:osCode       :" , oError:osCode
   ? "oError:severity     :" , oError:severity
   ? "oError:subCode      :" , oError:subCode
   ? "oError:subSystem    :" , oError:subSystem
   ? "oError:thread       :" , oError:thread
   ? "oError:tries        :" , oError:tries

   ? Replicate("-", 78)
   ? "CALLSTACK:"
   ? Replicate("-", 78)

   i:=nStackStart
   DO WHILE ! Empty(ProcName(++i))
      ? EHS_CALLED_FROM, Trim(ProcName(i))   + "(" + ;
                 LTrim(Str(ProcLine(i))) + ")"
   ENDDO
   ?

   SET ALTERNATE TO
   SET ALTERNATE OFF

   If AppType()<>APPTYPE_PM
      TONE(660,5)
      ? EHS_LOG_WRITTEN_TO(cErrorLog)
      If !Set(_SET_CONSOLE)
          OutErr(CHR(10)+CHR(13) + EHS_LOG_WRITTEN_TO(cErrorLog))
      EndIf
   Else
      MsgBox(EHS_LOG_WRITTEN_TO(cErrorLog))
   EndIf

   /* Restore previous settings */
   Set(_SET_PRINTER,   lPrint)
   Set(_SET_CONSOLE,   lConsole)
   Set(_SET_ALTFILE,   cAlternate)
   Set(_SET_ALTERNATE, lAlternate)
   Set(_SET_EXTRA,     lExtra)

Return



/* Split large line for Alert()-box output */
FUNCTION LineSplit(cMessage, nMaxCol)
LOCAL i
LOCAL cLines:=""
LOCAL nLines

   nLines:=MlCount(cMessage, nMaxCol,, .T.)
   FOR i:=1 TO nLines
        cLines += Rtrim(MemoLine(cMessage, nMaxCol, i,,.T.)) +";"
   NEXT
   If cLines[-1]==";"
      cLines:=Left(cLines,len(cLines)-1)
   EndIf
Return cLines

**************
Func ErrorText(nError)
**************
Local cRetVal:="", x:=0
Local aErrors:={{ 0 ,"No error"}, ;
                { 1 ,"Parameter has a wrong value"}, ;
                { 2 ,"Parameter has a wrong data type"}, ;
                { 3 ,"Missing parameter in function or method call"}, ;
                { 4 ,"Error in array index"}, ;
                {10 ,"Numeric overflow (number is too large or too small)"}, ;
                {11 ,"Division by zero"}, ;
                {12 ,"Invalid numeric value (e.g. 3 % 0 or float number is too large for array index)"}, ;
                {15 ,"Not enough memory or swapping space available"}, ;
                {16 ,"Stack overflow (e.g. with recursive calls. The stack can be dimensioned during linking with Link386)."}, ;
                {20 ,"Syntax error in macro expression"}, ;
                {21 ,"Function is not declared"}, ;
                {22 ,"Unknown variable. A variable name identifies neither a PRIVATE nor a PUBLIC variable"}, ;
                {23 ,"Object has no method with this name"}, ;
                {24 ,"Error in the METHOD..IS definition for a method: the message identifier is already used for another method and is not allowed for this method"}, ;
                {25 ,"Access to method is not allowed in this context (e.g. if a method is not declared as EXPORTED: it cannot be called within functions but only within methods)."}, ;
                {26 ,"Object has no member variable with this name"}, ;
                {27 ,"Error in the VAR..IS definition for a member variable: the message identifier is already used for another member variable or class name and is not allowed for this member variable"}, ;
                {28 ,"Access to member variable is not allowed in this context."}, ;
                {29 ,"Wrong data type for member variable"}, ;
                {30 ,"Access to instance method not allowed in class object"}, ;
                {31 ,"Access to instance variable not allowed in class object"}, ;
                {32 ,"Syntax error within class declaration"}, ;
                {33 ,"no instances allowed for this class: class contains DEFERRED methods or is destroyed"}, ;
                {34 ,"Class is already defined"}, ;
                {35 ,"Call of not implemented DEFERRED method"}, ;
                {36 ,"Method is FINAL and may not be overloaded"}, ;
                {37 ,"Class is not defined"}, ;
                {40 ,"Operating system error: this only occurs if the error cannot be identified otherwise."}, ;
                {41 ,"Internal data structures corrupted (an unexpected OS error occured)"}, ;
                {42 ,"XPF file has invalid version number"}, ;
                {43 ,"Attempt to save non persistent data. Data which cannot be made persistent are attempted to be saved via SAVE TO or Var2Bin()"}, ;
                {50 ,"DMLB sub-system is not initialized"}, ;
                {51 ,"Generic error in DMLB framework"}, ;
                {52 ,"Invalid DLL name for DBE (e.g. the file does not exist or does not have the .DLL extension)"}, ;
                {53 ,"DBE build name already in use"}, ;
                {54 ,"DBE build error"}, ;
                {55 ,"DBE component is defect"}, ;
                {56 ,"DBE initialization failed"}, ;
                {57 ,"DBE deinitialization failed"}, ;
                {58 ,"Operation is not supported by DBE"}, ;
                {59 ,"DBE private/specific error (error is caused by the DBE producer)"}, ;
                {60 ,"No default DBE available/defined"}, ;
                {61 ,"Invalid DBE name (DBE is not loaded or compound DBE is not built)"}, ;
                {62 ,"Invalid data type for database field"}, ;
                {63 ,"Length of database field was exceeded"}, ;
                {64 ,"Invalid length of index key (the NTXDBE supports max 256 characters)"}, ;
                {65 ,"Unknown symbol for database field"}, ;
                {66 ,"Unknown symbol for alias"}, ;
                {67 ,"Symbol is already in use"}, ;
                {68 ,"No table available/no database open"}, ;
                {69 ,"No order available/no index open"}, ;
                {70 ,"File cannot be created"}, ;
                {71 ,"File cannot be opened"}, ;
                {72 ,"Error while closing a file"}, ;
                {73 ,"Error while reading a file"}, ;
                {74 ,"Error while writing to a file"}, ;
                {75 ,"Operation requires a file to be exclusively opened"}, ;
                {76 ,"File lock is required for this operation"}, ;
                {77 ,"File is opened in read only mode"}, ;
                {78 ,"Auto lock error during APPEND BLANK"}, ;
                {79 ,"Record lock failed"}, ;
                {80 ,"Lost Update error was detected by DMLB (only in auto lock mode of a DBE)"}, ;
                {90 ,"Workarea does not fit into workspace"}, ;
                {100,"Operation is not allowed in this thread (e.g. when a database is opened in thread A it cannot be closed in thread B)"}, ;
                {101,"Error during printing or spooling"}, ;
                {102,"Method which must be overloaded has not been overloaded"}, ;
                {103,"The object being accessed is not registered (Most probably the :init() method has not been called)."}, ;
                {104,"The object (XBP) is in the wrong state to perform the operation (e.g. after :new() only :create() may be called but not :destroy())"}, ;
                {105,"Error while creating an object (XBP) (e.g. a non existing resource ID was assigned to the iVar :caption)"}, ;
                {130,"An invalid matrix was passed to a Gra..() operation (The matrix must be created using GraInitMatrix())."}, ;
                {133,"Application was not linked using /PM:PM-switch. The application was not linked with the correct link switch."}, ;
                {134,"Printer is not installed or is unavailable"}, ;
                {140,"Function is not supported"}, ;
                {141,"Error with dynamic loaded function"}, ;
                {142,"DLL couldn't be loaded"}, ;
                {143,"DLL was created with an incompatible version"}, ;
                {144,"unsupported calling convention"}, ;
                {145,"DLL must be compiled with /dll:dynamic"}, ;
                {333,"Function was terminated by BREAK()"}, ;
                {334,"Container has the wrong type"}, ;
                {335,"Write lock is not allowed for this string since a read lock has been set already"}, ;
                {336,"Read lock is not allowed for this string since a write lock has been set already"}, ;
                {337,"The limit for string locks has been reached for this container"}, ;
                {338,"String container is not locked"}, ;
                {339,"Handle is invalid or not in the correct state (locked/unlocked)"}}

Begin Sequence

   If (x:=Ascan(aErrors,{|e| e[1]=nError})) > 0
      cRetVal:=Ntrim(aErrors[x,1])+", "+aErrors[x,2]
   Else
      cRetVal:=Ntrim(nError)
   EndIf

End

Return cRetVal
