* Program..: Utilities.prg
* Author...: Jeremy Suiter
* Date.....: 25/03/2003
*
* Notice...: Copyright (c) 2003, Jeremy Suiter, All Rights Reserved
* Notes....: Visual DbEditor ToolTip functions(modifed from ASTAG sample)

#Include "Gra.ch"
#Include "Xbp.ch"
#Include "Appevent.ch"
#Include "Font.ch"

***************************
CLASS MagicHelp FROM Thread
***************************
   HIDDEN:
   METHOD DisplayToolTip()
   METHOD PaintTheTip()

   VAR lStop

   EXPORTED:
   VAR producerID
   VAR oLastMotionXBP
   VAR oLastTipXBP
   VAR oBlockedXBP
   VAR nLastTipTime
   VAR aLastMotionPos
   VAR oTip
   VAR lTipIsShown
   VAR nTipSensitivity

   INLINE METHOD init()
     ::nTipSensitivity:=1
     ::lStop := .F.
     ::thread:init()
     ::producerID:=ThreadID()
   Return

   EXPORTED:
   METHOD execute()
   METHOD atStart()
   METHOD showTip()
   METHOD hideTip()
   METHOD Stop()

ENDCLASS

************************
METHOD MagicHelp:atStart()
************************
   ::lTipIsShown      :=.F.
   ::oLastMotionXBP   :=NIL
   ::aLastMotionPos   :=NIL
Return

************************
METHOD MagicHelp:Execute()
************************
   Local nEvent, mp1:=0, mp2:=0, oXbp:=NIL
   Local nLastMotionTime:=0

#IfDef DBEDFUNCTION

   Do While OK2Loop      // JWL shut down thread on exit of main loop

#Else

   Do While !::lStop

#EndIf

     /*
      * Because our entire Event sniffer pools events, we have
      * to go sleep after each iteration otherwise we would
      * consume to much CPU resources for nothing!
      */
      Sleep(10)

     /*
      * here we go, using LastAppEvent to sniff into the event
      * queue of another thread
      */
      nEvent:=LastAppEvent(@mp1,@mp2,@oXbp,::producerID)

     /*
      * Because XbpIWindows are not in our interest in
      * terms of User-Related semantics we take the parent
      */
      If (oXbp:isDerivedFrom("XbpIWindow"))
         oXbp:=oXbp:setParent()
      EndIf

     /*
      * blocked XBP have input focus and have to be left out
      * in the tip heuristics
      */
      If (ValType(::oBlockedXBP)=="O" .And.;
         oXbp == ::oBlockedXBP)
         Loop
      EndIf

     /*
      * Only motion events are relevant to detect situations at which
      * we have to post a tip.
      */
      If (nEvent == xbeM_Motion)

     /*
      * Ok, a motion event has occured, and it was not over the
      * XBP which is blocked - means of input-focus - so we
      * remove the block
      */
         ::oBlockedXBP:=NIL

     /*
      * We check here If the motion was again over the same XBP as the
      * last motion event was, then we validate for the same mouse pos.
      * If that's all the case we only have to ensure that the
      * ::nTipSensitivity timeframe was reached, If so we post the tip.
      */
      If (ValType(::oLastMotionXBP) == "O" .And. oXbp == ::oLastMotionXBP)

         If (::aLastMotionPos[1] == mp1[1] .And. ;
             ::aLastMotionPos[2] == mp1[2])

            If ((Seconds() - nLastMotionTime) > ::nTipSensitivity)
               ::showTip()
            EndIf

         Else
            ::aLastMotionPos:=AClone(mp1)
            nLastMotionTime:=Seconds()
         EndIf

     /*
      * Check If the current XBP we are over has the same parent
      * as the XBP for which we have posted the last TIP. In that
      * case we check out If the timeframe between these two actions
      * was under half a second. Are all prev. conditions fullfilled
      * we assume the user is interested in another tip for the XBP
      * in the neighborhood of its current - so it goes.
      */
      ElseIf (ValType(::oLastTipXBP) == "O" .And. ;
         oXbp:setParent() == ::oLastTipXBP:setParent() .And. ;
         oXbp != ::oLastTipXBP)

         ::hideTip()
         ::oLastMotionXBP :=oXbp
         ::aLastMotionPos :=AClone(mp1)
         nLastMotionTime:=Seconds()

         If (::nLastTipTime>0 .And. ;
           (Seconds()-::nLastTipTime)<=0.5)
            ::showTip()
         EndIf

      /*
       * Ok, nothing special has occured, so we simply store the
       * event data related to the motion to be used in the next
       * pass.
       */
       Else
          ::oLastMotionXBP :=oXbp
          ::aLastMotionPos :=AClone(mp1)
          nLastMotionTime  :=Seconds()
          ::hideTip()
       EndIf

     /*
      * Any event except PaintEvent has to hide the Tip. In
      * addition, we mark the XBP as blocked - because
      * user-interaction via Keyboard or Mouse has taken
      * place - to avoid Tip postings while the user
      * interacts with the XbasePART
      */
      ElseIf nEvent != xbeP_Paint

         If (oXbp==::oLastTipXBP)
           ::oBlockedXBP   :=oXbp
         EndIf

         ::oLastMotionXBP:=NIL
         ::hideTip()
      EndIf

   EndDo

Return

************************
METHOD MagicHelp:Stop()
************************
   ::lStop := .T.
   ::Synchronize(0)
RETURN NIL

************************
METHOD MagicHelp:showTip()
************************

Begin Sequence

   // no help label so don't display
   If !(ValType(::oLastMotionXBP:helpLink)=="O" .And. ::oLastMotionXBP:helpLink:isDerivedFrom("MagicHelpLabel"))
       Break
   EndIf

   If (!::lTipIsShown)
      ::DisplayToolTip(::oLastMotionXBP)
      ::oTip:show()
      ::lTipIsShown :=.T.
      ::oLastTipXBP :=::oLastMotionXBP
      ::oBlockedXBP:=NIL
   EndIf

End

Return

************************
METHOD MagicHelp:hideTip()
************************

   If (::lTipIsShown)
      ::oTip:hide()
      ::oTip:destroy()
      ::lTipIsShown:=.F.
      ::nLastTipTime:=Seconds()
   EndIf

Return

*******************************
METHOD MagicHelp:DisplayToolTip(oXbpRequestingHint)
*******************************
   Local cText:=""
   Local aPos

  /*
   * calculate absolute position of motion event and adjust it
   * about the mouse pointer size
   */
   aPos :=calcAbsolutePosition(::aLastMotionPos,oXbpRequestingHint)
   aPos[1] += 6
   aPos[2] -= 30

  /*
   * Check If the XBP for which we have to post a hint has an
   * associated Helplabel, and If so, retrieve the ID and
   * hint-message from the database
   */
   If (ValType(oXbpRequestingHint:helpLink)=="O" .And. ;
      oXbpRequestingHint:helpLink:isDerivedFrom("MagicHelpLabel"))
      cText:=oXbpRequestingHint:helpLink:getID()
   EndIf

  /*
   * Ok, now lets paint the TIP
   */
   ::oTip:=XbpStatic():new()
   ::oTip:options:=XBPSTATIC_TYPE_FGNDFRAME
   ::oTip:create(AppDesktop(),AppDesktop(), aPos, { 0 , 0 })
   ::PaintTheTip(cText)

Return(Self)

****************************
METHOD MagicHelp:PaintTheTip(cText)
****************************
   Local aAreaAttr, aStringAttr
   Local oPS
   Local aPoints
   Local aSize:={0,0}

   oPS:=::oTip:lockPS()
   aPoints:=GraQueryTextBox(oPS, cText)
   ::oTip:unlockPS()

   aSize[1]:=(aPoints[3,1] - aPoints[1,1]) + 8
   aSize[2]:=(aPoints[1,2] - aPoints[2,2]) + 4

   ::oTip:setSize(aSize,.F.)
   oPS:=::oTip:lockPS()

   aAreaAttr:=Array(GRA_AA_COUNT)
   aAreaAttr [ GRA_AA_COLOR ]:=XBPSYSCLR_INFOBACKGROUND
   GraSetAttrArea(oPS, aAreaAttr)
   GraBox(oPS, { 0, 0}, {aSize[1]-1, aSize[2]-1}, GRA_OUTLINEFILL)

   aStringAttr:=Array(GRA_AS_COUNT)
   aStringAttr [ GRA_AS_COLOR ]:=XBPSYSCLR_INFOTEXT
   GraSetAttrString(oPS, aStringAttr)
   GraStringAt(oPS, {4,4}, cText)

   ::oTip:unLockPS(oPS)
Return(Self)

**************************************
CLASS MagicHelpLabel FROM XbpHelpLabel
**************************************
   VAR cID

   INLINE METHOD init(cID)
      ::cID:=cID
   Return

   EXPORTED:
   INLINE METHOD getID()
   Return(::cID)

ENDCLASS


/*
 * This function calculates the absolute position
 * from a given position relative to an XbasePART
 */
********************************
Static Func CalcAbsolutePosition(aPos,oXbp)
********************************
Local aAbsPos:=AClone(aPos)
Local oParent:=oXbp
Local oDesktop:=AppDesktop()

   Do While oParent<>oDesktop
      aAbsPos[1] += oParent:currentPos()[1]
      aAbsPos[2] += oParent:currentPos()[2]
      oParent:=oParent:setParent()
   EndDo

Return(aAbsPos)
