* Program..: XbpGetC.prg
* Author...: Jeremy Suiter
* Date.....: 19/03/2003
*
* Notice...: Copyright (c) 2003, Jeremy Suiter, All Rights Reserved
* Notes....: Visual DbEditor get controller functions (modified from ATSAG samples)

#include "Appevent.ch"
#include "Common.ch"
#include "Get.ch"
#include "Xbp.ch"


#define  CB_CLOSE                 1
#define  CB_SETINPUTFOCUS         2
#define  CB_SETDISPLAYFOCUS       3
#define  CB_KILLDISPLAYFOCUS      4
#define  CB_KEYBOARD              5
#define  CB_TABACTIVATE           6

#define  CB_GETSET_COUNT          6

#define  SV_GET_CB                1
#define  SV_XBP                   2
#define  SV_OLD_CB                3

#define  SV_COUNT                 3


CLASS XbpGetController
   PROTECTED:
   VAR aControls
   VAR aGetList
   VAR aCallback
   VAR aGetSet
   VAR aNewCB
   VAR nCurrent
   VAR oWindow

   VAR thisXbp, lastXbp
   VAR readExit

   METHOD collect
   METHOD saveCallBack
   METHOD getCallback
   METHOD restCallBack
   METHOD activate

   EXPORTED:
                             ** These must be exported
   VAR doPostValidate        // enable validation
   VAR exitState READONLY    // last exit state
   VAR setHome               // XbpGet cursor must be set to first position

   METHOD init, create, destroy

                             ** Methods reserved for XbpDialog/XbpCrt windows
   METHOD setDisplayFocus    // window obtains display focus
   METHOD killDisplayFocus   // window looses display focus
   METHOD setInputFocus      // window obtains input focus
   METHOD close              // window is about to be closed

                             ** Methods reserved for other XBPs
   METHOD addEditControl     // Adds an editable XBP
   METHOD addFocusControl    // Adds an XBP that can get input focus
   METHOD tabActivate        // intercepts Tab activation events
   METHOD keyboard           // keyboard navigation for XBPs contained in window
                             ** Methods for navigation
   METHOD first              // go to first control
   METHOD last               // go to last control
   METHOD up                 // go to previous control
   METHOD down               // go to next control
                             ** Methods for editing
   METHOD preValidate        // data validation on single entry field level
   METHOD postValidate       //
   METHOD clearEventLoop     // clear event loop after validation
   METHOD read               // start editing
   METHOD killRead           // enforce edit termination
   METHOD getActive          // Returns current edit control
   METHOD getList            // array containing edit controls
   METHOD setData            // transfer data into edit controls
   METHOD getData            // get data from edit controls
ENDCLASS


METHOD XbpGetController:init(oWindow)
   ::oWindow       :=oWindow
   ::doPostValidate:=.T.
   ::setHome       :=.F.
Return self


/*
 * Create new callback code blocks and get/set codeblocks to replace
 * existing callback code blocks.
 */
METHOD XbpGetController:create(oWindow)
   ::aControls:={}
   ::aGetList :={}
   ::aCallback:={}
   ::nCurrent :=1

   If Valtype(oWindow) == "O"
      ::oWindow:=oWindow
   EndIf

   ::aGetSet                     :=Array(CB_GETSET_COUNT)
   ::aGetSet[CB_CLOSE           ]:={|o,b| IIf(PCount()<>2, o:close           , o:close           :=b) }
   ::aGetSet[CB_SETINPUTFOCUS   ]:={|o,b| IIf(PCount()<>2, o:setInputFocus   , o:setInputFocus   :=b) }
   ::aGetSet[CB_SETDISPLAYFOCUS ]:={|o,b| IIf(PCount()<>2, o:setDisplayFocus , o:setDisplayFocus :=b) }
   ::aGetSet[CB_KILLDISPLAYFOCUS]:={|o,b| IIf(PCount()<>2, o:killDisplayFocus, o:killDisplayFocus:=b) }
   ::aGetSet[CB_KEYBOARD        ]:={|o,b| IIf(PCount()<>2, o:keyboard        , o:keyboard        :=b) }
   ::aGetSet[CB_TABACTIVATE     ]:={|o,b| IIf(PCount()<>2, o:tabActivate     , o:tabActivate     :=b) }

   ::aNewCB                      :=Array(CB_GETSET_COUNT)
   ::aNewCB [CB_CLOSE           ]:={|mp1,mp2,o| ::close           (mp1, mp2, o) }
   ::aNewCB [CB_SETINPUTFOCUS   ]:={|mp1,mp2,o| ::setInputFocus   (mp1, mp2, o) }
   ::aNewCB [CB_SETDISPLAYFOCUS ]:={|mp1,mp2,o| ::setDisplayFocus (mp1, mp2, o) }
   ::aNewCB [CB_KILLDISPLAYFOCUS]:={|mp1,mp2,o| ::killDisplayFocus(mp1, mp2, o) }
   ::aNewCB [CB_KEYBOARD        ]:={|mp1,mp2,o| ::keyboard        (mp1, mp2, o) }
   ::aNewCB [CB_TABACTIVATE     ]:={|mp1,mp2,o| ::tabActivate     (mp1, mp2, o) }

   ::collect(::oWindow)
Return self


/*
 * Restore the original state in the window and release all arrays.
 */
METHOD XbpGetController:destroy

   /*
    * Remove self's reference
    */
   AEval(::aGetList, {|o| IIf(IsMemberVar(o, "Controller"), o:Controller:=NIL, NIL) })

   /*
    * Restore changed callback codeblocks
    */
   ::restCallback()

   ::aCallback:=;
   ::aControls:=;
   ::aGetList :=;
   ::aGetSet  :=;
   ::aNewCB   :=;
   ::thisXbp  :=;
   ::lastXbp  :=;
   ::oWindow  :=NIL
Return self


/*
 * Save and replace all callback code blocks for keyboard and
 * focus change events. This enables the controller to take corrective
 * action when a pre- or postvalidation fails.
 */
METHOD XbpGetController:collect(oXbp)
   LOCAL aXbp, i, imax

   If oXbp:isDerivedFrom("XbpDialog") .OR. ;
      oXbp:isDerivedFrom("XbpCrt"   )

      ::saveCallback(oXbp, ::aGetSet[CB_CLOSE           ], ::aNewCB[CB_CLOSE           ])
      ::saveCallback(oXbp, ::aGetSet[CB_SETINPUTFOCUS   ], ::aNewCB[CB_SETINPUTFOCUS   ])
      ::saveCallback(oXbp, ::aGetSet[CB_KILLDISPLAYFOCUS], ::aNewCB[CB_KILLDISPLAYFOCUS])
      ::saveCallback(oXbp, ::aGetSet[CB_SETDISPLAYFOCUS ], ::aNewCB[CB_SETDISPLAYFOCUS ])
   EndIf

   If oXbp:isDerivedFrom("XbpDialog")
      Return ::collect(oXbp:drawingArea)
   EndIf

   /*
    * ::aGetList collects editable Xbase Parts.
    * ::aControls contains Xbase Parts which can get focus in the window.
    */

   If IsMemberVar(oXbp, "dataLink")
      ::addEditControl(oXbp)

      If IsMemberVar(oXbp, "tabStop")
         ::addFocusControl(oXbp)
      EndIf
   Else
      If oXbp:isDerivedFrom("XbpTabpage")
         ::saveCallback(oXbp, ::aGetSet[CB_TABACTIVATE], ::aNewCB[CB_TABACTIVATE])

      ElseIf IsMemberVar(oXbp, "tabStop")
         ::addFocusControl(oXbp)

      EndIf

      aXbp:=oXbp:childList()
      imax:=Len(aXbp)
      FOR i:=1 TO imax
         ::collect(aXbp[i])
      NEXT
   EndIf

   If IsMemberVar(oXbp, "Controller")
      oXbp:controller:=self
      If Atail(::aGetList) <> oXbp
         AAdd(::aGetList, oXbp)
      EndIf
   EndIf

Return self


/*
 * ::aGetList collects editable Xbase Parts.
 */
METHOD XbpGetController:addEditControl(oXbp)
   If Ascan(::aGetList, oXbp) == 0
      AAdd(::aGetList, oXbp)
      If IsMemberVar(oXbp, "tabStop")
         oXbp:tabStop:=.T.
      EndIf
   EndIf
Return self



/*
 * ::aControls contains Xbase Parts which can get focus in the window.
 * The :tabStop flag defines whether or not an XBP is added to that list
 * (keyboard navigation within a window).
 */
METHOD XbpGetController:addFocusControl(oXbp)

   If oXbp:tabStop
      AAdd(::aControls, oXbp)

      If oXbp:isDerivedFrom("XbpGet")
        /*
         * XbpGet does not require a :keyboard code block, since its
         * :keyboard() method handles the communication with XbpGetController.
         */
         ::saveCallback(oXbp, ::aGetSet[CB_KEYBOARD], NIL)
      Else
        /*
         * Replace and save the :keyboard code block to ensure that a
         * focus change is handled by XbpGetController.
         */
         ::saveCallback(oXbp, ::aGetSet[CB_KEYBOARD], ::aNewCB[CB_KEYBOARD])
      EndIf
   EndIf

Return self



/*
 * Save and replace one callback code block.
 */
METHOD XbpGetController:saveCallBack(oXbp, bGetSet, bNew)
   LOCAL bOld:=Eval(bGetSet, oXbp)
   LOCAL aSave[ SV_COUNT ]

   aSave[SV_GET_CB]:=bGetSet
   aSave[SV_XBP   ]:=oXbp
   aSave[SV_OLD_CB]:=bOld

   EVal(bGetSet, oXbp, bNew)
   AAdd(::aCallback, aSave)
Return self


/*
 * Return a saved callback code block.
 */
METHOD XbpGetController:getCallback(oXbp, bGetSet)
   LOCAL i:=Ascan(::aCallback, {|a| a[SV_XBP] == oXbp .AND. a[SV_GET_CB] == bGetSet })
Return IIf(i==0, NIL, ::aCallback[i, SV_OLD_CB])


/*
 * Restore all saved callback code blocks.
 */
METHOD XbpGetController:restCallBack
   AEval(::aCallback, {|a| Eval(a[SV_GET_CB], a[SV_XBP], a[SV_OLD_CB]) })
Return self


/*
 * Window gains input focus.
 * Set focus to current editable XBP inside window and enable validation
 */
METHOD XbpGetController:setInputFocus
   LOCAL bBlock

   ::doPostValidate:=.T.

   If ::lastXbp <> NIL
      ::thisXbp:=::lastXbp
      Return SetAppFocus(::lastXbp)
   EndIf

   If ::thisXbp == NIL
      ::thisXbp:=::aControls[ ::nCurrent ]
   EndIf

   bBlock:=::getCallBack(::oWindow, ::aGetSet[ CB_SETINPUTFOCUS ])
   If bBlock <> NIL
      /*
       * Evaluate the original :setInputFocus codeblock.
       */
      Eval(bBlock, NIL, NIL, ::oWindow )
   EndIf

   SetAppFocus(::thisXbp)
Return self


/*
 * Window gains focus.
 * Set focus to current editable XBP inside window and enable validation
 */
METHOD XbpGetController:setDisplayFocus
   LOCAL bBlock:=::getCallBack(::oWindow, ::aGetSet[ CB_SETDISPLAYFOCUS ])

   If bBlock <> NIL
      /*
       * Evaluate the original :setDisplayFocus codeblock.
       */
      Eval(bBlock, NIL, NIL, ::oWindow )
   EndIf

Return ::setInputFocus()


/*
 * Window looses focus.
 * Keep track of XBP that has focus and disable validation
 */
METHOD XbpGetController:killDisplayFocus
   LOCAL bBlock
   LOCAL oXbp:=SetAppFocus()
   LOCAL i   :=Ascan(::aControls, {|o| o == oXbp })

   If i > 0 .AND. i <> ::nCurrent
      ::nCurrent:=i
      ::thisXbp:=::aControls[i]
   EndIf

   ::doPostValidate:=.F.
   ::lastXbp       :=::thisXbp
   ::thisXbp       :=NIL
   ::exitState     :=GE_LASTGET

   bBlock:=::getCallBack(::oWindow, ::aGetSet[ CB_KILLDISPLAYFOCUS ])
   If bBlock <> NIL
      /*
       * Evaluate the original :killDisplayFocus codeblock.
       */
      Eval(bBlock, NIL, NIL, ::oWindow )
   EndIf
Return self


/*
 * Window is about to be closed.
 */
METHOD XbpGetController:close
   LOCAL lSave:=::doPostValidate
   LOCAL bBlock

   If ::thisXbp == NIL
      ::thisXbp:=::lastXbp
   EndIf

   /*
    * Enforce post validation on current XbpGet when window gets closed
    */
   ::doPostValidate:=.T.
   ::readExit      :=::postValidate()
   ::doPostValidate:=lSave

   If ::readExit
      /*
       * Window can be closed. Evaluate the original :close code block
       */
      bBlock:=::getCallBack(::oWindow, ::aGetSet[ CB_CLOSE ])

      If bBlock <> NIL
         Eval(bBlock, NIL, NIL, ::oWindow )
      EndIf
      ::destroy()
   Else
      /*
       * Post validation failed. Keep window open.
       */
      MsgBox("Illegal data. Please correct before closing the window.")

      If ::oWindow:getFrameState() == XBPDLG_FRAMESTAT_MINIMIZED
         ::oWindow:setFrameState(XBPDLG_FRAMESTAT_NORMALIZED)
      EndIf

      /*
       * Void the xbeP_Close event and set focus to XBP whose validation failed.
       */
      SetAppFocus(::oWindow)
      SetAppFocus(::lastXbp)
   EndIf
Return self


/*
 * A TabPage is clicked.
 */
METHOD XbpGetController:tabActivate(mp1, mp2, oTab)
   LOCAL bBlock:=::getCallback(oTab, ::aGetSet[ CB_TABACTIVATE ])

   If ::exitState <> GE_LASTGET
      /*
       * Postvalidation is successful.
       * Evaluate the original :tabActivate code block.
       */
      oTab:=Eval(bBlock, mp1, mp2, oTab)
   Else
      /*
       * Postvalidation failed.
       * TabPage remains in the foreground.
       */
      oTab:=::thisXbp:setParent()
      DO WHILE ! oTab:isDerivedfrom("XbpTabpage")
         oTab:=oTab:setParent()
         If oTab == ::oWindow .OR. oTab:isDerivedFrom("XbpIWindow")
            EXIT
         EndIf
      ENDDO
      oTab:toFront()
   EndIf
Return oTab


/*
 * Keyboard handling
 */
METHOD XbpGetController:keyboard(nKey, mp2, oXbp)
   LOCAL bBlock:=SetAppEvent(nKey)
   LOCAL i

   mp2 := mp2

   ::exitState:=GE_NOEXIT
   ::setHome  :=.T.

   DO CASE
   CASE bBlock <> NIL
     /*
      * A code block is associated with an event.
      */

   CASE nKey == xbeK_ALT_F4
      PostAppEvent(xbeP_Close,,, ::oWindow)
      Return .T.

   CASE nKey == xbeK_TAB
      ::exitState:=GE_DOWN
      bBlock     :={|| ::down() }

   CASE nKey == xbeK_SH_TAB
      ::exitState:=GE_UP
      bBlock     :={|| ::up() }

   CASE ::thisXbp <> NIL .AND. .NOT. Empty(::thisXbp:childList())
     /*
      * NOTE: XBP contains other controls and must handle the
      *       following keys on its own
      */

   CASE nKey == xbeK_ENTER
      If ::thisXbp:isDerivedFrom("XbpPushbutton")
         PostAppEvent(xbeP_Activate, NIL, NIL, ::thisXbp)
         ::setHome  :=.F.
      Else
         ::exitState:=GE_DOWN
         bBlock     :={|| ::down() }
      EndIf

   CASE nKey == xbeK_DOWN
      ::exitState:=GE_DOWN
      bBlock     :={|| ::down() }

   CASE nKey == xbeK_UP
      ::exitState:=GE_UP
      bBlock     :={|| ::up() }

   CASE nKey == xbeK_CTRL_HOME
      ::exitState:=GE_TOP
      bBlock     :={|| ::first() }

   CASE nKey == xbeK_CTRL_END
      ::exitState:=GE_BOTTOM
      bBlock     :={|| ::last() }

   OTHERWISE
      ::setHome  :=.F.
      bBlock     :=::getCallback(::thisXbp, ::aGetSet[CB_KEYBOARD])

   ENDCASE

   If bBlock <> NIL
      i:=Ascan(::aControls, {|o| o == oXbp })
      If i > 0 .AND. i <> ::nCurrent
         /*
          * A non-XbpGet object has received input focus via mouse click
          */
         ::nCurrent:=i
         ::lastXbp :=::thisXbp
         ::thisXbp :=::aControls[i]
      EndIf
      /*
       * Evaluate the code block associated with a key
       */
      Eval(bBlock, nKey, NIL, ::thisXbp)
   EndIf
Return (bBlock <> NIL)


/*
 * Validation before an Xbase Part gains input focus.
 */
METHOD XbpGetController:preValidate(oXbp)
   LOCAL lValid:=.T.

   DEFAULT oXbp TO ::thisXbp

   If ::lastXbp == oXbp .OR. ::exitState == GE_LASTGET
      /*
       * No prevalidation is required when:
       *  - the window that contains ::lastXbp has lost focus and gets it
       *    it back now. The prevalidation is done already in that case.
       *  - a postvalidation has failed and focus must be set back to
       *    the XbpGet holding illegal data.
       */
   ElseIf IsMethod(oXbp, "preValidate")
      /*
       * Focus change between controls in the same window
       * requires a prevalidation.
       */
      lValid:=oXbp:preValidate()

   EndIf

   If lValid
      If ::exitState == GE_LASTGET .AND. ::lastXbp <> NIL
         /*
          * Focus goes back to XbpGet where postvalidation failed
          */
         SetAppFocus(::lastXbp)
         lValid     :=.F.
         ::clearEventLoop()
         ::exitState:=GE_NOEXIT
      Else
         ::thisXbp  :=;
         ::lastXbp  :=oXbp
         ::nCurrent :=AScan(::aControls, oXbp)
      EndIf
   Else
      /*
       * Prevalidation failed.
       */
      Tone(500)
      ::setHome  :=.T.

      If ::exitState == GE_UP
         ::up()
      ElseIf ::exitState == GE_DOWN .OR. ::lastXbp == NIL
         ::down()
      Else
         SetAppFocus(::lastXbp)
      EndIf
      ::exitState:=GE_WHEN
   EndIf
Return lValid


/*
 * Validation before an Xbase Part looses input focus.
 */
METHOD XbpGetController:postValidate(oXbp)
   LOCAL lValid:=.T., lSave

   DEFAULT oXbp TO ::thisXbp

   If .NOT. ::doPostValidate     .OR. ;
       oXbp        == NIL        .OR. ;
       ::exitState == GE_WHEN    .OR. ;
       ::exitState == GE_LASTGET .OR. ;
       (::oWindow:getFrameState() == XBPDLG_FRAMESTAT_MINIMIZED .AND. LastAppEvent() <> xbeP_Close)
      /*
       * Postvalidation must Return .T. when:
       *  - no XBP has got focus in the window yet (initial display)
       *  - the window containing the XBP looses focus (another window is clicked)
       *  - focus is taken away due to a failed prevalidation
       *  - the window is minimized
       */
   ElseIf IsMethod(oXbp, "postValidate")
      /*
       * Prevent from recursive postvalidation when the input focus is
       * lost due to a MsgBox() displayed in the postvalidation routine.
       */
      lSave           :=::doPostvalidate
      ::doPostvalidate:=.F.

      If .NOT. oXbp:postValidate()
         Tone(1000)
         lValid     :=.F.
         ::lastXbp  :=oXbp
         ::exitState:=GE_LASTGET

         If oXbp == ::thisXbp
            SetAppFocus(oXbp)
         ElseIf oXbp == ::lastXbp
            ::thisXbp :=oXbp
            ::nCurrent:=AScan(::aControls, oXbp)
         EndIf
      EndIf
      ::doPostvalidate:=lSave
   EndIf

   If ::exitState == GE_NOEXIT
      ::exitState:=GE_WRITE
   EndIf
Return lValid


/*
 * Process pending events after a failed Pre validation.
 */
METHOD XbpGetController:clearEventLoop
   LOCAL nEvent, mp1:=Nil, mp2:=Nil, oXbp:=Nil

   DO WHILE NextAppEvent() <> xbe_None
      nEvent:=AppEvent(@mp1, @mp2, @oXbp)
      oXbp:handleEvent(nEvent, mp1, mp2)
   ENDDO
Return self


/*
 * Sets or retrives the current Xbase Part.
 */
METHOD XbpGetController:getActive(oXbp)
   LOCAL oOld:=IIF(::thisXbp == NIL, ::lastXbp, ::thisXbp)
   LOCAL i

   If oOld == NIL
      oOld:=SetAppFocus()
   EndIf

   If oXbp <> NIL .AND. (i:=AScan(::aGetList, oXbp)) > 0
      ::nCurrent:=AScan(::aControls, ::aGetList[i])
   EndIf
Return oOld


/*
 * Methods for navigation.
 */
METHOD XbpGetController:first
   ::nCurrent:=1

   If .NOT. ::aControls[ ::nCurrent ]:isVisible()
      ::down()
   EndIf
Return ::Activate()


METHOD XbpGetController:last
   ::nCurrent:=Len(::aControls)

   If .NOT. ::aControls[ ::nCurrent ]:isVisible()
      ::up()
   EndIf
Return ::Activate()


METHOD XbpGetController:up
   LOCAL lExit:=.F.

   DO WHILE ! lExit
      If -- ::nCurrent < 1
         ::nCurrent:=Len(::aControls)
      EndIf
      lExit:=::aControls[ ::nCurrent ]:isVisible()
   ENDDO
Return ::Activate()


METHOD XbpGetController:down
   LOCAL lExit:=.F.

   DO WHILE ! lExit
      If ++ ::nCurrent > Len(::aControls)
         ::nCurrent:=1
      EndIf
      lExit:=::aControls[ ::nCurrent ]:isVisible()
   ENDDO
Return ::Activate()


/*
 * Set input focus to the current Xbase Part.
 */
METHOD XbpGetController:activate
   If ! Empty(::aControls)
      ::thisXbp:=::aControls[ ::nCurrent ]
      SetAppFocus(::thisXbp)
   EndIf
Return ::thisXbp


/*
 * Modal (default) or modeless read method.
 */
METHOD XbpGetController:read(nPos, lModal)
   LOCAL nEvent, mp1:=Nil, mp2:=Nil, oXbp:=Nil, oFocus
   LOCAL oDlg:=::owindow

   DEFAULT lModal TO .T., ;
           nPos   TO  1

   oDlg:show()
   ::setData()
   oFocus:=::getActive(::aGetList[ nPos ])

   If lModal
      oDlg:setModalState(XBP_DISP_APPMODAL)

      ::readExit:=.F.

      ::activate()
      DO WHILE .NOT. ::readExit
         nEvent:=AppEvent(@mp1, @mp2, @oXbp)
         oXbp:handleEvent(nEvent, mp1, mp2)
         if nEvent     == xbeP_Keyboard .and.;
            mp1        == xbeK_ESC
            ::readExit := .t.
         endif         
      ENDDO

      oDlg:setModalState(XBP_DISP_MODELESS)
      SetAppFocus(oFocus)
   Else
      ::activate()
   EndIf
Return self


/*
 * The :close() method performs a postvalidation. If this is successful,
 * :readExit is set to .T. and the event loop in :read() is exited.
 */
METHOD XbpGetController:killRead
   If ::oWindow <> NIL
      ::close(::oWindow)
   Else
      ::readExit:=.T.
   EndIf
Return self


METHOD XbpGetController:getList
Return AClone(::aGetList)


METHOD XbpGetController:setData
   AEval(::aGetlist, {|o| IIf(o:dataLink==NIL, NIL, o:setData()) })
Return self


METHOD XbpGetController:getData
   AEval(::aGetlist, {|o| o:getData() })
Return self
